/*
 * Copyright (c) y2k Micha Moskal <malekith@topnet.pl>
 *
 * This file is part of cryptoFTP - cftp protocol implementation.
 * CryptoFTP is free software. You can redistribute it and/or modify it under
 * the terms of the GNU General Public License (see COPYING.GPL for details).
 *
 * $Id: client.c,v 1.9 2000/05/05 22:33:10 malekith Exp $
 */

#include "config.h"
#include "cftp.h"
#include "utils.h"
#include "version.h"

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/time.h>

#ifdef USE_READLINE
#ifdef HAVE_READLINE_READLINE_H
#include <readline/readline.h>
#elif defined (HAVE_READLINE_H)
#include <readline.h>
#endif
#ifdef HAVE_READLINE_HISTORY_H
#include <readline/history.h>
#elif defined (HAVE_HISTORY_H)
#include <history.h>
#endif /* HAVE_READLINE_HISTORY_H */
#endif /* USE_READLINE */

static const char *attention =
"\n"
"     00000000000000000000000000000000000000000000000000000000000\n"
"     00                  W  A  R  N  I  N  G                  00\n"
"     00000000000000000000000000000000000000000000000000000000000\n"
"\n"
"host identification for %s has changed. it's possible that\n"
"someone at the other end of the link is pretending that he's %s.\n"
"it's also possible that system admin @ %s has just changed his\n"
"host's id.\n"
"to avoid this messsage, please edit your host file (~/.cftp/known_hosts)\n"
"continuing connection could lead to compromise of your password\n";

char *x_get_pass();
void do_command(char *line, char *p);

char *user, *host, *home;
char *def_user;
int port;
char *cipher_to_use = DEFAULT_CIPHER;

struct clib_state *ctx;

void close_link()
{
	if (ctx) {
		clib_free(ctx);
		printf("<--- connection closed ]]]\n");
	}
	ctx = 0;
}

#ifndef USE_READLINE
char *readline(const char *prompt)
{
	char line[1000];
	
	printf("%s", prompt);
	fflush(stdout);
	if (!fgets(line, sizeof line, stdin))
		return 0;
	*(strchr(line, 0) - 1) = 0;
	return xstrdup(line);
}

void add_history(const char *line)
{
	(void)line;
}

void init_readline()
{
}
#endif /* !USE_READLINE */

#ifdef USE_READLINE
#ifdef COMPLETION
char *compl_gen(char *text, int state);
char *compl_cmd(char *text, int state);
char **cftp_completion(char *text, int start, int end)
{
	(void)end;
	if (start == 0)
		return completion_matches(text, compl_cmd);
	else if ((*rl_line_buffer|0x20) == 'l' || (*rl_line_buffer|0x20) == 'p' ||
	    ((*rl_line_buffer|0x20) == 'z' && (*rl_line_buffer|0x20) == 'p'))
		return 0;
	else
		return completion_matches(text, compl_gen);
}
#endif
void init_readline()
{
	rl_readline_name = "cftp";
#ifdef COMPLETION
	rl_attempted_completion_function = cftp_completion;
#endif
}
#endif /* USE_READLINE */

int abort_please = 0;

void sig_int(int s)
{
	abort_please = 1;
	signal(s, sig_int);
}

/* connection died */
void cdie(const char *str)
{
/*
	error("reseting connection due the external problem. "
		"you gonna have to reconnect\n"
		"problem was: '%s'", str);
*/
	printf("error: %s\n", str);
	close_link();
}

void kill_children()
{
	signal(SIGHUP, SIG_IGN);
	kill(-getpgrp(), SIGHUP);
	signal(SIGHUP, SIG_DFL);
}

static void sig_handler(int sig)
{
	switch (sig) {
	case SIGUSR1:
		signal(SIGHUP, sig_handler);
		break;
	case SIGUSR2:
		signal(SIGHUP, SIG_DFL);
		break;
	}
	signal(sig, sig_handler);
}

void setup_signals()
{
	/* this ought to be ignored in any event */
	signal(SIGTSTP, sig_handler); 
	/* this will cause forwarding process to ignore SIGHUP sent when 
	 * parent exits */
	signal(SIGUSR1, sig_handler);
	/* reset SIGHUP to its default deadly behavior */ 
	signal(SIGUSR2, sig_handler);
}

void init()
{
	sig_int(SIGINT);
	init_readline();
	
	if (getenv("CONFIG_DIR"))
		home = xprintf("%s/cftp", getenv("CONFIG_DIR"));
	else if (getenv("HOME"))
		home = xprintf("%s/.cftp", getenv("HOME"));
	else
		home = xprintf(ETC_DIR);
	
	if (access(home, R_OK) != 0) {
		printf("created home dir (%s)\n", home);
		mkdir(home, 0700);
	}
		
	clib_init(home);
	
	atexit(kill_children);
	atexit(clib_close);
}

static int ask_question()
{
	char *l;
	int r = -1;

	while (r == -1 &&
	       (l = readline("are you sure you want to connect "
			     "(yes/no): "))) {
		if (!l)
			break;
		if (strcmp(l, "yes") == 0)
			r = 1;
		if (strcmp(l, "no") == 0)
			r = 0;
		xfree(l);
	}
	if (r == -1) {
		printf("ctrl-d pressed\n");
		exit(0);
	}
	return r;
}

struct clib_state *create_ctx(const char *host, int port)
{
	struct clib_state *ctx;
	int err;
	
	ctx = clib_alloc();

	printf("connecting to %s : %d ... ", host, port);
	err = clib_open1(ctx, host, port);
	if (err)
		goto shit;
	printf("ok\nwaiting for RSA key ... ");
	err = clib_open2(ctx, cipher_to_use);
	if (err < 0)
		goto shit;
	printf("ok\n");
	if (err == 1) {
		printf("new host\n%s not found in list of known hosts "
		       "(%s/known_hosts)\n", host, home);
		if (ask_question() != 1)
			goto bye;
	} else if (err == 2) {
		printf(attention, host, host, host);
		if (ask_question() != 1)
			goto bye;
	} else if (err)
		goto shit;

	err = clib_open3(ctx);
	
	if (err)
		goto shit;
		
	printf("<<< %s\n", ctx->error);
	clib_clr_error(ctx);

	return ctx;

shit:
	if (ctx->error)
		printf("error: %s\n", ctx->error);
bye:
	clib_free(ctx);
	return 0;
}

void create_link()
{
	int err;
	char *pass;

	close_link();
	
	ctx = create_ctx(host, port);
	
	if (!ctx) {
		printf(">! connection not created\n");
		return;
	}
	
	if (!user) {
		if (!def_user)
			def_user = xstrdup(getenv("USER") ?
					   getenv("USER") : "user");
		printf("login@%s (%s) : ", host, def_user);
		user = readline("");

		if (!user || user[0] == 0) {
			xfree(user);
			user = xstrdup(def_user);
		}
	}
	printf("%s's password: ", user);
	pass = x_get_pass();
	if (!pass) {
		clib_free(ctx);
		ctx = 0;
		return;
	}
	err = clib_open4(ctx, user, pass);
	xfree0(pass, strlen(pass));

	printf("%s%s\n", err == 0 ? "<<< connected. message:\n" : 
	       "!!! access denied: ", ctx->message);

	if (err) {
		clib_free(ctx);
		ctx = 0;
	}
	
	if (user) {
		xfree(def_user);
		def_user = xstrdup(user);
	}
	xfree(user);
	user = 0;
}

int forward(struct clib_state *ctx, char *p)
{
	int p1,p2,pid;
	char *r, *target;
	
	while (isspace(*p))
		p++;

	if (strchr(p, '^') && strchr(strchr(p, '^') + 1, '^')) {
		/* LPORT^rhost^RPORT */
		p1 = atoi(p); p = strchr(p, '^') + 1;
		target = p; p = strchr(p, '^');
		*p++ = 0;
		p2 = atoi(p);
	} else if (strchr(p, '^')) {
		target = p;
		p1 = atoi(p);
		while (*p != '^' && isdigit(*p))
			p++;
		if (*p == '^') {
			/* LPORT^RPORT */
			target = "localhost";
			p2 = atoi(++p);
		} else {
			/* rhost^RPORT */
			p = strchr(p, '^');
			*p++ = 0;
			p1 = p2 = atoi(++p);
		}
	} else if (strchr(p, ':') && strchr(strchr(p, ':') + 1, ':')) {
		/* LPORT:rhost:RPORT */
		p1 = atoi(p); p = strchr(p, ':') + 1;
		target = p; p = strrchr(p, ':'); /* note use of strRchr */
		*p++ = 0;
		p2 = atoi(p);
	} else if (strchr(p, ':')) {
		target = p;
		p1 = atoi(p);
		while (*p != ':' && isdigit(*p))
			p++;
		if (*p == ':') {
			/* LPORT:RPORT */
			target = "localhost";
			p2 = atoi(++p);
		} else {
			/* rhost:RPORT */
			p = strchr(p, '^');
			*p++ = 0;
			p1 = p2 = atoi(++p);
		}
	} else {
		target = "localhost";
		p1 = p2 = atoi(p);
	}
	
	printf("]]] requesting fwd { %s:%d -  - ---> %d }\n", target, p2, p1);
	if (clib_forward(ctx, p2, target, p1)) {
		printf("!!! fwd failed: %s\n", ctx->error);
		clib_clr_error(ctx);
		return -1;
	}
	printf("]]] forking child ...\n");
	pid = fork();
	if (pid == -1) {
		perror("fork");
		return -1;
	}
	if (pid) {
		printf("]]] got child [%d]\n", pid);
		return 0;
	}
	setup_signals();
	r = clib_forward_forked(ctx);
	if (r)
		printf("\n]]] fwd: %s\n",r);
	xfree(r);
	exit(0);
}

int split_hp(char *p);

int anonfwd(char *p)
{
	char *h;
	struct clib_state *ctx;
	
	while (isspace(*p))
		p++;
	h = p;
	while (!isspace(*p) && *p)
		p++;
		
	if (!*p) {
		printf("errro: no forward target\n");
		return -1;
	}
	
	*p++ = 0;
	while (isspace(*p))
		p++;
	
	ctx = create_ctx(h, split_hp(h));
	if (!ctx)
		return -1;

	return forward(ctx, p);
}

int main(int argc, char **argv)
{
	char *line;

	setbuf(stdout, 0);
	init();
	
	random_mark();
	
	argv++,argc--;
	
	if (argc > 1 && strcmp(*argv, "-f")==0) {
		argv++,argc--;
		if (anonfwd(*argv))
			exit(1);
		printf("\n"
"IiI you'll now be droped in cftp command line. you may connect to some other\n"
"IiI host as well as use Ctrl-Z to suspend cftp and use your established\n"
"IiI forward link\n\n");
	} else if (argc && (*argv)[0] == '-' && 
		   ((*argv)[1] == 'v' || 
		   	((*argv)[1] == '-' && (*argv)[2] == 'v'))) {
		printf("%s\n", SOFTWARE_VERSION);
		exit(0);
	} else if (argc && **argv == '-') {
		printf("usage:\n"
"	cftp 				or\n"
"	cftp [user@]host[:port]		or\n"
"	cftp [user@]host[^port]		or\n"
"	cftp --version			or\n"
"	cftp -f 'host[:port] [local_port:][remote_host:]remote_port' or\n"
"	cftp -f 'host[^port] [local_port^][remote_host^]remote_port'\n\n"
"'^' can be used in place of ':' to allow using IPv6 addresses.\n");
		exit(1);
	} else if (argc) {
		do_command("open", *argv);
	}
	
	while ((line = readline("cftp> "))) {
		char *p;
		
		random_mark();
		
		if (!*line) { xfree(line); continue; }
		add_history(line);
		p = line;
		while (*p && *p != ' ')
			p++;
		while (*p == ' ')
			*p++ = 0;
		while (*p && *(strchr(p, 0) - 1) == ' ')
			*(strchr(p, 0) - 1) = 0;
			
		do_command(line, p);
		
		if (ctx && clib_clr_error(ctx)) {
			printf("- -=-> connection is no longer usable\n");
			close_link();
		}
			
		xfree(line);
		
		random_mark();
	}
	return 0;
}
