/*
 * Copyright (c) y2k Micha Moskal <malekith@topnet.pl>
 *
 * This file is part of libcftp - cftp protocol implementation.
 * CryptoFTP is free software. You can redistribute it and/or modify it under
 * the terms of the GNU Library General Public License (see COPYING.LGPL 
 * for details).
 *
 * $Id: crypto.c,v 1.4 2000/05/05 22:20:26 malekith Exp $
 */

#include "config.h"
#include "utils.h"
#include "rsa.h"
#include "cipher.h"

#ifdef __linux__
#include <fcntl.h>
#include <unistd.h>
#endif


void crypto_init()
{
	rsa_set_memory_alloc();
#ifdef __linux__
	{
		int fd;
		byte buf[1000];
		
		fd = open("/dev/urandom", O_RDONLY);
		if (fd != -1) {
			read(fd, buf, sizeof(buf));
			random_add(buf, sizeof(buf));
			close(fd);
		}
	}
#else /* ! __linux__ */
	{
		int n;
		double d;
		n = 100;
		while (n--) {
			d = dtime();
			random_add(&d, sizeof(d));
		}
		/* will be anyway initialised later with ~/.cftp/rand_seed */
	}
#endif /* __linux__ */
	init_ciphers();
}

#if defined(__unix__) || defined(__CYGWIN__)
#include <time.h>
#include <sys/time.h>
double dtime()
{
	struct timeval tv;
	gettimeofday(&tv, 0);
	return (double)tv.tv_sec + (double)tv.tv_usec / 1000000;
}
#else /* ! __unix__ */
#ifdef WINDA
double dtime()
{
	FILETIME tm;
	GetSystemTimeAsFileTime(&tm);
	return ((__int64)tm.dwHighDateTime<<32|tm.dwLowDateTime)/10000000.0;
}
#else
double dtime()
{
#error "write dtime() !"
}
#endif /* WINDA */
#endif /* __unix__ */

