/*
 * Copyright (c) y2k Micha Moskal <malekith@topnet.pl>
 *
 * This file is part of cryptoFTP - cftp protocol implementation.
 * CryptoFTP is free software. You can redistribute it and/or modify it under
 * the terms of the GNU General Public License (see COPYING.GPL for details).
 *
 * $Id: keygen.c,v 1.2 2000/04/23 16:39:03 malekith Exp $
 */

#include "config.h"

#include <sys/types.h>
#include <stdio.h>
#include <termios.h>
#include <unistd.h>
#include <sys/stat.h>
#include <ctype.h>
#include <stdlib.h>

#include "rsa.h"
#include "utils.h"

const char *rsa_private_key_name = RSA_PRIVATE_KEY_NAME;
const char *rsa_public_key_name = RSA_PUBLIC_KEY_NAME;

#define die(a) do { fprintf(stderr, a); exit(1); } while(0)

void write_key(struct rsa_key *key)
{
	FILE *f;
	
	f = fopen(rsa_public_key_name, "w");
	if (!f)
		die("can't open public key");
	fprintf(f, "cftp public key version 1.0\n%d\n", key->bits);
	mpz_out_str(f, 16, key->n); fputc('\n', f);
	mpz_out_str(f, 16, key->e); fputc('\n', f);
	fclose(f);

	umask(077);
	f = fopen(rsa_private_key_name, "w");
	if (!f)
		die("can't open private key");
	fprintf(f, "cftp private key version 1.0\n%d\n", key->bits);
	mpz_out_str(f, 16, key->n); fputc('\n', f);
	mpz_out_str(f, 16, key->e); fputc('\n', f);
	mpz_out_str(f, 16, key->d); fputc('\n', f);
	/*
	mpz_out_str(f, 16, key->u); fputc('\n', f);
	mpz_out_str(f, 16, key->p); fputc('\n', f);
	mpz_out_str(f, 16, key->q); fputc('\n', f);
	*/
	fclose(f);
}

void seed_random()
{
	struct termios tattr, tattr_s;
	int n;
	char c, c1 = 0;
	double last = 0;
	
	if (!isatty(0))
		die("stdin is NOT terminal");

	fprintf(stderr, "please press random keys. please *don't* pick one and HOLD it !\n" 
			"i'll use delays between characters to generate random bits for key\n"
			"two lines of dots must show up\n");
		
	tcgetattr(0, &tattr_s);
	tcgetattr(0, &tattr);
	tattr.c_lflag &= ~(ECHO|ICANON);
	tattr.c_cc[VMIN] = 1;
	tattr.c_cc[VTIME] = 0;
	tcsetattr(0, TCSAFLUSH, &tattr);
	
	n = 159;
	while (n) {
		random_mark();
		if (dtime() - last > 0.2 || c != c1) {
			n--;
			fprintf(stderr, ".");
		}
		c1 = c;
		last = dtime();
		read(0, &c, 1);
		random_mark();
	}
	fprintf(stderr, "\nenough, enough, now please press [enter] key to begin.\n");
	
	while (c != '\n' && c != '\r')
		read(0, &c, 1);
	
	tcsetattr(0, TCSAFLUSH, &tattr_s);
}
extern int rsa_verbose;
int main(int argc, char **argv)
{
	struct rsa_key key;
	int bits = 1024;
	(void)argc;
	crypto_init();
	
	argv++;

/* this is safe to do only on linux, where /dev/random is available and was
 * already used */
#ifdef __linux__	
	if (*argv && strcmp(*argv, "-s") == 0) {
		seed_random();
		argv++;
	}
#else
	seed_random();
#endif
	
	if (*argv && isdigit(*argv[0]))
		bits = atoi(*argv);
	
	if (bits < 768) {
		fprintf(stderr, "cannot generate so small key. falling back to 1024 bits\n");
		bits = 1024;
	}
		
	rsa_verbose = 1;
	
	rsa_generate_key(&key, bits);
	write_key(&key);
	rsa_destroy_key(&key);
	return 0;
}
